-- Orthotropic membrane, Green strain

with
  text_io,
  ada.integer_text_io,
  ada.long_float_text_io,
  ada.numerics.long_elementary_functions,
  terms,
  nlp;

procedure orthomem1 is

  subtype fmat3x1 is terms.fmat1 (1 .. 3);

  subtype fmat9x1 is terms.fmat1 (1 .. 9);

  type cst_rec is
    record
      c    : long_float;
      s    : long_float;
      p    : long_float;
      area : long_float;
      l1   : long_float;
      l2   : long_float;
      l3   : long_float;
      glob : terms.imat1 (1 .. 3);
      u1   : terms.fmat1 (1 .. 3);
      u2   : terms.fmat1 (1 .. 3);
      u3   : terms.fmat1 (1 .. 3);
      h    : terms.fmat2 (1 .. 3, 1 .. 3);
    end record;

  type cst_vec is array (integer range <>) of cst_rec;

  type cst_ptr is access cst_vec;

  halt                                    : exception;
  optimal                                 : boolean;
  nn,ne,n,maxls,minor,maxminor,maxlist    : integer;
  major,maxmajor                          : integer;
  lstol,mntol,mnref,maxnorm,f             : long_float;
  young1,young2,poisson12,shear12,density : long_float;
  index                                   : terms.imat2_ptr;
  xo,yo,zo,x,g                            : terms.fmat1_ptr;
  known,action,force                      : terms.fmat2_ptr;
  cst                                     : cst_ptr;

  procedure scan (io : in text_io.file_type) is

    ascii : character;

  begin
    loop
      text_io.get(io,ascii);
      text_io.skip_line(io);
      exit when ascii = ':';
    end loop;
  end scan;

  procedure input is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io;

    e,glob,axis,nk : integer;
    io             : file_type;

  begin
    open(io, in_file, "input.txt");

    -- The iterations terminate if the infinity norm of the
    -- gradient becomes less than or equal to MAXNORM, which is
    -- calculated as:

    -- MAXNORM = MNTOL * MNREF

    -- If MNTOL is specified as a non-positive value, it is
    -- replaced by 0.001.

    -- If MNREF is specified as a non-positive value, it is
    -- replaced by the infinite norm of the gradient, evaluated at
    -- the starting point.

    scan(io);
    get(io,lstol);
    get(io,maxls);
    get(io,mntol);
    get(io,mnref);
    get(io,maxminor);
    get(io,maxlist);
    get(io,maxmajor);
    skip_line(io);

    if mntol <= 0.0 then
      mntol := 0.001;
    end if;

    -- young1    : Young's modulus in direction 1, multiplied by thickness
    -- young2    : Young's modulus in direction 2, multiplied by thickness
    -- poisson12 : Poissons ratio for strain in direction 2 when stressed in direction 1
    -- shear12   : Shear modulus in the 1,2 plane, multiplied by thickness
    -- density   : force per unit area in the Z direction

    scan(io);
    get(io,young1);
    get(io,young2);
    get(io,poisson12);
    get(io,shear12);
    get(io,density);
    skip_line(io);

    -- nodes
    scan(io);
    get(io,nn);
    skip_line(io);

    -- initiate xo, yo, zo
    xo := new terms.fmat1 (1 .. nn);
    yo := new terms.fmat1 (1 .. nn);
    zo := new terms.fmat1 (1 .. nn);

    skip_line(io);
    for k in 1 .. nn loop
      get(io,glob);
      get(io,xo(glob));
      get(io,yo(glob));
      get(io,zo(glob));
      skip_line(io);
    end loop;

    -- elements
    scan(io);
    get(io,ne);
    skip_line(io);

    -- initiate cst
    cst := new cst_vec (1 .. ne);

    skip_line(io);
    for k in 1 .. ne loop
      get(io,e);
      get(io,cst(e).glob(1));
      get(io,cst(e).glob(2));
      get(io,cst(e).glob(3));
      get(io,cst(e).c);
      get(io,cst(e).s);
      skip_line(io);
    end loop;

    -- initiate index
    index := new terms.imat2 (1 .. nn, 1 .. 3);
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        index(glob,axis) := 0;
      end loop;
    end loop;

    -- initiate known
    known := new terms.fmat2 (1 .. nn, 1 .. 3);
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        known(glob,axis) := 0.0;
      end loop;
    end loop;

    -- known displacements
    scan(io);
    get(io,nk);
    skip_line(io);

    if nk > 0 then
      skip_line(io);
      for k in 1 .. nk loop
        get(io,glob);
        get(io,axis);
        get(io,known(glob,axis));
        skip_line(io);
        index(glob,axis) := 1;
      end loop;
    end if;

    -- initiate action
    action := new terms.fmat2 (1 .. nn, 1 .. 3);
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        action(glob,axis) := 0.0;
      end loop;
    end loop;

    -- nodal actions
    scan(io);
    get(io,nk);
    skip_line(io);

    if nk > 0 then
      skip_line(io);
      for k in 1 .. nk loop
        get(io,glob);
        get(io,axis);
        get(io,action(glob,axis));
        skip_line(io);
      end loop;
    end if;

    -- pressure on membrane elements
    scan(io);
    get(io,nk);
    skip_line(io);

    if nk > 0 then
      skip_line(io);
      for k in 1 .. nk loop
        get(io,e);
        get(io,cst(e).p);
        skip_line(io);
      end loop;
    end if;

    close(io);
  end input;

  procedure cstobj (cst            : in cst_rec;
                    st1,st2,st3,fe : out long_float;
                    x              : in terms.fmat1_ptr;
                    ge             : out fmat9x1) is

    glob,i                      : integer;
    delta1,e1,e2,e3,sv1,sv2,sv3 : long_float;
    z                           : terms.fmat1 (1 .. 3);
    de1,de2,de3                 : terms.fmat1 (1 .. 9);
    xe                          : terms.fmat2 (1 .. 3, 1 .. 3);

  begin
    -- global displacements
    for loc in 1 .. 3 loop
      glob := cst.glob(loc);
      for axis in 1 .. 3 loop
        i := index(glob,axis);
        if i > n then
          xe(loc,axis) := known(glob,axis);
        else
          xe(loc,axis) := x(i);
        end if;
      end loop;
    end loop;

    -- side 1

    -- z
    for axis in 1 .. 3 loop
      z(axis) := (xe(3,axis) - xe(2,axis)) / cst.l1;
    end loop;

    -- delta + 1.0
    delta1 := 0.0;
    for axis in 1 .. 3 loop
      delta1 := delta1 + (cst.u1(axis) + z(axis)) ** 2;
    end loop;

    -- e1
    e1 := (delta1 - 1.0) / 2.0;

    -- de1
    de1(1) := 0.0;
    de1(2) := 0.0;
    de1(3) := 0.0;

    de1(7) := (cst.u1(1) + z(1)) / cst.l1;
    de1(8) := (cst.u1(2) + z(2)) / cst.l1;
    de1(9) := (cst.u1(3) + z(3)) / cst.l1;

    de1(4) := -de1(7);
    de1(5) := -de1(8);
    de1(6) := -de1(9);

    -- side 2

    -- z
    for axis in 1 .. 3 loop
      z(axis) := (xe(1,axis) - xe(3,axis)) / cst.l2;
    end loop;

    -- delta + 1.0
    delta1 := 0.0;
    for axis in 1 .. 3 loop
      delta1 := delta1 + (cst.u2(axis) + z(axis)) ** 2;
    end loop;

    -- e2
    e2 := (delta1 - 1.0) / 2.0;

    -- de2
    de2(4) := 0.0;
    de2(5) := 0.0;
    de2(6) := 0.0;

    de2(1) := (cst.u2(1) + z(1)) / cst.l2;
    de2(2) := (cst.u2(2) + z(2)) / cst.l2;
    de2(3) := (cst.u2(3) + z(3)) / cst.l2;

    de2(7) := -de2(1);
    de2(8) := -de2(2);
    de2(9) := -de2(3);

    -- side 3

    -- z
    for axis in 1 .. 3 loop
      z(axis) := (xe(2,axis) - xe(1,axis)) / cst.l3;
    end loop;

    -- delta + 1.0
    delta1 := 0.0;
    for axis in 1 .. 3 loop
      delta1 := delta1 + (cst.u3(axis) + z(axis)) ** 2;
    end loop;

    -- e3
    e3 := (delta1 - 1.0) / 2.0;

    -- de3
    de3(7) := 0.0;
    de3(8) := 0.0;
    de3(9) := 0.0;

    de3(4) := (cst.u3(1) + z(1)) / cst.l3;
    de3(5) := (cst.u3(2) + z(2)) / cst.l3;
    de3(6) := (cst.u3(3) + z(3)) / cst.l3;

    de3(1) := -de3(4);
    de3(2) := -de3(5);
    de3(3) := -de3(6);

    -- stress * thickness
    st1 := cst.h(1,1) * e1 + cst.h(1,2) * e2 + cst.h(1,3) * e3;
    st2 := cst.h(2,1) * e1 + cst.h(2,2) * e2 + cst.h(2,3) * e3;
    st3 := cst.h(3,1) * e1 + cst.h(3,2) * e2 + cst.h(3,3) * e3;

    -- stress * volume
    sv1 := st1 * cst.area;
    sv2 := st2 * cst.area;
    sv3 := st3 * cst.area;

    -- residue
    for i in 1 .. 9 loop
      ge(i) := sv1 * de1(i) + sv2 * de2(i) + sv3 * de3(i);
    end loop;

    -- energy
    fe := (sv1 * e1 + sv2 * e2 + sv3 * e3) / 2.0;
  end cstobj;

  procedure objective (n : in integer;
                       f : out long_float;
                       x : in terms.fmat1_ptr;
                       g : in terms.fmat1_ptr) is

    glob,i,s       : integer;
    st1,st2,st3,fe : long_float;
    ge             : fmat9x1;

  begin
    -- initiate objective, gradient
    f := 0.0;
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        i := index(glob,axis);
        if i <= n then
          f := f - force(glob,axis) * x(i);
          g(i) := -force(glob,axis);
        end if;
      end loop;
    end loop;

    -- elements
    for e in 1 .. ne loop
      cstobj(cst(e),st1,st2,st3,fe,x,ge);

      -- update objective
      f := f + fe;

      -- update gradient
      s := 0;
      for loc in 1 .. 3 loop
        glob := cst(e).glob(loc);
        for axis in 1 .. 3 loop
          s := s + 1;
          i := index(glob,axis);
          if i <= n then
            g(i) := g(i) + ge(s);
          end if;
        end loop;
      end loop;
    end loop;
  end objective;

  package qn is new nlp.qnewton(objective);

  procedure start is

    use
      ada.numerics.long_elementary_functions;

    sqrt2 : constant long_float := sqrt(2.0);

    h11,h12,h22,h33,poisson21           : long_float;
    glob1,glob2,glob3,s1,s2             : integer;
    cos1,cos2,sin1,sin2,sin3,norm       : long_float;
    a11,a12,a13,a21,a22,a23,a31,a32,a33 : long_float;
    b11,b12,b13,b21,b22,b23,b31,b32,b33 : long_float;
    c21,c22,c23,c31,c32,c33             : long_float;
    t11,t12,t13,t33                     : long_float;
    u                                   : terms.fmat1 (1 .. 3);

  begin
    -- number of unknowns
    n := 3 * nn;
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        if index(glob,axis) = 1 then
          n := n - 1;
        end if;
      end loop;
    end loop;

    -- new sequence
    s1 := 0;
    s2 := n;
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        if index(glob,axis) = 0 then
          s1 := s1 + 1;
          index(glob,axis) := s1;
        else
          s2 := s2 + 1;
          index(glob,axis) := s2;
        end if;
      end loop;
    end loop;

    -- matrix H^
    poisson21 := young2 / young1 * poisson12;
    h11 := young1 / (1.0 - poisson12 * poisson21);
    h12 := poisson12 * young2 / (1.0 - poisson12 * poisson21);
    h22 := young2 / (1.0 - poisson12 * poisson21);
    h33 := shear12 + shear12;

    -- elements
    for e in 1 .. ne loop
      glob1 := cst(e).glob(1);
      glob2 := cst(e).glob(2);
      glob3 := cst(e).glob(3);

      -- u1
      u(1) := xo(glob3) - xo(glob2);
      u(2) := yo(glob3) - yo(glob2);
      u(3) := zo(glob3) - zo(glob2);
      norm := sqrt(u(1) ** 2 + u(2) ** 2 + u(3) ** 2);
      for axis in 1 .. 3 loop
        cst(e).u1(axis) := u(axis) / norm;
      end loop;
      cst(e).l1 := norm;

      -- u2
      u(1) := xo(glob1) - xo(glob3);
      u(2) := yo(glob1) - yo(glob3);
      u(3) := zo(glob1) - zo(glob3);
      norm := sqrt(u(1) ** 2 + u(2) ** 2 + u(3) ** 2);
      for axis in 1 .. 3 loop
        cst(e).u2(axis) := u(axis) / norm;
      end loop;
      cst(e).l2 := norm;

      -- u3
      u(1) := xo(glob2) - xo(glob1);
      u(2) := yo(glob2) - yo(glob1);
      u(3) := zo(glob2) - zo(glob1);
      norm := sqrt(u(1) ** 2 + u(2) ** 2 + u(3) ** 2);
      for axis in 1 .. 3 loop
        cst(e).u3(axis) := u(axis) / norm;
      end loop;
      cst(e).l3 := norm;

      -- cosine alpha1
      cos1 := 0.0;
      for axis in 1 .. 3 loop
        cos1 := cos1 - cst(e).u2(axis) * cst(e).u3(axis);
      end loop;

      if cos1 > 1.0 then
        cos1 := 1.0;
      end if;

      -- cosine alpha2
      cos2 := 0.0;
      for axis in 1 .. 3 loop
        cos2 := cos2 - cst(e).u3(axis) * cst(e).u1(axis);
      end loop;

      if cos2 > 1.0 then
        cos2 := 1.0;
      end if;

      -- sine alpha1
      sin1 := sqrt(1.0 - cos1 ** 2);

      -- sine alpha2
      sin2 := sqrt(1.0 - cos2 ** 2);

      -- sine alpha3
      sin3 := sin1 * cos2 + sin2 * cos1;

      -- inverse of matrix C
      c21 := cos1 / sin2 / sin3;
      c22 := cos2 / sin1 / sin3;
      c23 := -cos1 * cos2 / sin1 / sin2;
      c31 := -sin1 / sin2 / sin3 / sqrt2;
      c32 := sin2 / sin1 / sin3 / sqrt2;
      c33 := (cos2 / sin2 - cos1 / sin1) / sqrt2;

      -- matrix T
      t11 := cst(e).c ** 2;
      t12 := cst(e).s ** 2;
      t13 := cst(e).c * cst(e).s * sqrt2;
      t33 := t11 - t12;

      -- A = Matrix T * inverse of matrix C
      a11 := t12 * c21 + t13 * c31;
      a12 := t12 * c22 + t13 * c32;
      a13 := t12 * c23 + t13 * c33 + t11;
      a21 := t11 * c21 - t13 * c31;
      a22 := t11 * c22 - t13 * c32;
      a23 := t11 * c23 - t13 * c33 + t12;
      a31 := t13 * c21 + t33 * c31;
      a32 := t13 * c22 + t33 * c32;
      a33 := t13 * c23 + t33 * c33 - t13;

      -- B = matrix H^ * matrix A
      b11 := a11 * h11 + a21 * h12;
      b12 := a12 * h11 + a22 * h12;
      b13 := a13 * h11 + a23 * h12;
      b21 := a11 * h12 + a21 * h22;
      b22 := a12 * h12 + a22 * h22;
      b23 := a13 * h12 + a23 * h22;
      b31 := a31 * h33;
      b32 := a32 * h33;
      b33 := a33 * h33;

      -- H = transpose of matrix A * matrix B
      cst(e).h(1,1) := a11 * b11 + a21 * b21 + a31 * b31;
      cst(e).h(1,2) := a11 * b12 + a21 * b22 + a31 * b32;
      cst(e).h(1,3) := a11 * b13 + a21 * b23 + a31 * b33;
      cst(e).h(2,2) := a12 * b12 + a22 * b22 + a32 * b32;
      cst(e).h(2,3) := a12 * b13 + a22 * b23 + a32 * b33;
      cst(e).h(3,3) := a13 * b13 + a23 * b23 + a33 * b33;
      -- symmetry
      cst(e).h(2,1) := cst(e).h(1,2);
      cst(e).h(3,1) := cst(e).h(1,3);
      cst(e).h(3,2) := cst(e).h(2,3);

      -- area
      cst(e).area := cst(e).l1 * cst(e).l2 * sin3 / 2.0;
    end loop;

    -- initiate x
    x := new terms.fmat1 (1 .. n);
    for k in 1 .. n loop
      x(k) := 0.0;
    end loop;

    -- initiate g
    g := new terms.fmat1 (1 .. n);
    for k in 1 .. n loop
      g(k) := 0.0;
    end loop;

    -- initiate force
    force := new terms.fmat2 (1 .. nn, 1 .. 3);
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        force(glob,axis) := 0.0;
      end loop;
    end loop;
  end start;

  procedure vertex (glob : in integer;
                    vx   : out fmat3x1) is

    i : integer;

  begin
    vx(1) := xo(glob);
    vx(2) := yo(glob);
    vx(3) := zo(glob);

    for axis in 1 .. 3 loop
      i := index(glob,axis);
      if i <= n then
        vx(axis) := vx(axis) + x(i);
      else
        vx(axis) := vx(axis) + known(glob,axis);
      end if;
    end loop;
  end vertex;

  procedure autocad is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io;

    vx : fmat3x1;
    io : file_type;

  begin
    create(io, out_file, "autocad.scr");

    -- set running object snap mode to NONE
    put(io,"osmode 0");
    new_line(io);

    -- undeformed
    put(io,"layer make Undeformed ");
    new_line(io);

    -- create a polygon mesh
    put(io,"pface");
    new_line(io);

    -- vertex location
    for glob in 1 .. nn loop
      put(io,xo(glob),1,8,3);
      put(io,",");
      put(io,yo(glob),1,8,3);
      put(io,",");
      put(io,zo(glob),1,8,3);
      new_line(io);
    end loop;
    new_line(io);

    -- define faces
    for e in 1 .. ne loop
      put(io,cst(e).glob(1),0);
      put(io," ");
      put(io,cst(e).glob(2),0);
      put(io," ");
      put(io,cst(e).glob(3),0);
      new_line(io);
      new_line(io);
    end loop;
    new_line(io);

    -- deformed
    put(io,"layer make Deformed ");
    new_line(io);

    -- create a polygon mesh
    put(io,"pface");
    new_line(io);

    -- vertex location
    for glob in 1 .. nn loop
      vertex(glob,vx);
      put(io,vx(1),1,8,3);
      put(io,",");
      put(io,vx(2),1,8,3);
      put(io,",");
      put(io,vx(3),1,8,3);
      new_line(io);
    end loop;
    new_line(io);

    -- define faces
    for e in 1 .. ne loop
      put(io,cst(e).glob(1),0);
      put(io," ");
      put(io,cst(e).glob(2),0);
      put(io," ");
      put(io,cst(e).glob(3),0);
      new_line(io);
      new_line(io);
    end loop;
    new_line(io);

    put(io,"layer off Undeformed ");
    new_line(io);

    put(io,"layer set 0 ");
    new_line(io);

    put(io,"zoom ext ext");
    new_line(io);

    close(io);
  end autocad;

  procedure setforce is

    glob        : integer;
    aux         : long_float;
    vx1,vx2,vx3 : fmat3x1;
    u,v,w       : terms.fmat1 (1 .. 3);

  begin
    -- initiate force
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        force(glob,axis) := 0.0;
      end loop;
    end loop;

    -- add applied actions
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        force(glob,axis) := force(glob,axis) + action(glob,axis);
      end loop;
    end loop;

    -- add self weight
    for e in 1 .. ne loop
      aux := cst(e).area * density / 3.0;
      for loc in 1 .. 3 loop
        glob := cst(e).glob(loc);
        force(glob,3) := force(glob,3) - aux;
      end loop;
    end loop;

    -- add pressure
    for e in 1 .. ne loop
      glob := cst(e).glob(1);
      vertex(glob,vx1);

      glob := cst(e).glob(2);
      vertex(glob,vx2);

      glob := cst(e).glob(3);
      vertex(glob,vx3);

      -- u, v
      for axis in 1 .. 3 loop
        u(axis) := vx3(axis) - vx2(axis);
        v(axis) := vx1(axis) - vx3(axis);
      end loop;

      -- w = u x v
      w(1) := u(2) * v(3) - u(3) * v(2);
      w(2) := u(3) * v(1) - u(1) * v(3);
      w(3) := u(1) * v(2) - u(2) * v(1);

      -- pressure
      aux := cst(e).p / 6.0;
      for axis in 1 .. 3 loop
        w(axis) := aux * w(axis);
      end loop;

      for loc in 1 .. 3 loop
        glob := cst(e).glob(loc);
        for axis in 1 .. 3 loop
          force(glob,axis) := force(glob,axis) + w(axis);
        end loop;
      end loop;
    end loop;
  end setforce;

  procedure putchar (io : in text_io.file_type;
                     c  : in character;
                     n  : in integer) is

  begin
    for k in 1 .. n loop
      text_io.put(io,c);
    end loop;
  end putchar;

  procedure output is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io,
      ada.numerics.long_elementary_functions;

    glob,i,last,s        : integer;
    aux,st1,st2,st3,fe   : long_float;
    s1,s2,cos1,cos2,cos3 : long_float;
    vx                   : fmat3x1;
    ge                   : fmat9x1;
    io                   : file_type;

  begin
    create(io, out_file, "output.txt");

    putchar(io,'-',65);
    new_line(io);

    put(io,"| Major Iterations                            | ");
    put(io,major,15);
    put(io," |");
    new_line(io);

    put(io,"| Last Minor Iterations                       | ");
    put(io,minor,15);
    put(io," |");
    new_line(io);

    put(io,"| Total Potential Energy                      | ");
    put(io,f,2,8,3);
    put(io," |");
    new_line(io);

    put(io,"| Allowable Maximum Residue                   | ");
    put(io,maxnorm,2,8,3);
    put(io," |");
    new_line(io);

    putchar(io,'-',65);
    new_line(io);

    -- displacements
    new_line(io);
    new_line(io);
    putchar(io,'-',65);
    new_line(io);
    put(io," Node             Displ X             Displ Y             Displ Z");
    new_line(io);
    putchar(io,'-',65);
    new_line(io);

    new_line(io);

    for glob in 1 .. nn loop
      put(io,glob,5);
      for axis in 1 .. 3 loop
        putchar(io,' ',5);
        i := index(glob,axis);
        if i > n then
          put(io,known(glob,axis),2,8,3);
        else
          put(io,x(i),2,8,3);
        end if;
      end loop;
      new_line(io);
    end loop;

    -- vertex location
    new_line(io);
    new_line(io);
    putchar(io,'-',65);
    new_line(io);
    put(io," Node             Coord X             Coord Y             Coord Z");
    new_line(io);
    putchar(io,'-',65);
    new_line(io);

    new_line(io);

    for glob in 1 .. nn loop
      vertex(glob,vx);
      put(io,glob,5);
      for axis in 1 .. 3 loop
        putchar(io,' ',5);
        put(io,vx(axis),2,8,3);
      end loop;
      new_line(io);
    end loop;

    -- initiate support reactions
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        force(glob,axis) := -force(glob,axis);
      end loop;
    end loop;

    -- principal stresses
    new_line(io);
    new_line(io);
    putchar(io,'-',47);
    new_line(io);
    put(io,"Element            Stress-1            Stress-2");
    new_line(io);
    putchar(io,'-',47);
    new_line(io);

    new_line(io);

    for e in 1 .. ne loop
      cstobj(cst(e),st1,st2,st3,fe,x,ge);

      -- cosine alpha1
      cos1 := 0.0;
      for axis in 1 .. 3 loop
        cos1 := cos1 - cst(e).u2(axis) * cst(e).u3(axis);
      end loop;

      if cos1 > 1.0 then
        cos1 := 1.0;
      end if;

      -- cosine alpha2
      cos2 := 0.0;
      for axis in 1 .. 3 loop
        cos2 := cos2 - cst(e).u3(axis) * cst(e).u1(axis);
      end loop;

      if cos2 > 1.0 then
        cos2 := 1.0;
      end if;

      -- cosine alpha3
      cos3 := 0.0;
      for axis in 1 .. 3 loop
        cos3 := cos3 - cst(e).u1(axis) * cst(e).u2(axis);
      end loop;

      if cos3 > 1.0 then
        cos3 := 1.0;
      end if;

      -- principal stresses
      aux := 0.0;
      aux := aux + st1 * st2 * (cos3 ** 2 - 0.5);
      aux := aux + st1 * st3 * (cos2 ** 2 - 0.5);
      aux := aux + st2 * st3 * (cos1 ** 2 - 0.5);
      aux := 4.0 * aux + st1 ** 2 + st2 ** 2 + st3 ** 2;
      aux := sqrt(aux);

      s1 := ((st1 + st2 + st3) + aux) / 2.0;
      s2 := ((st1 + st2 + st3) - aux) / 2.0;

      put(io,e,7);
      putchar(io,' ',5);
      put(io,s1,2,8,3);
      putchar(io,' ',5);
      put(io,s2,2,8,3);
      new_line(io);

      -- update support reactions
      s := 0;
      for loc in 1 .. 3 loop
        glob := cst(e).glob(loc);
        for axis in 1 .. 3 loop
          s := s + 1;
          force(glob,axis) := force(glob,axis) + ge(s);
        end loop;
      end loop;
    end loop;

    -- support reactions
    new_line(io);
    new_line(io);
    putchar(io,'-',65);
    new_line(io);
    put(io," Node             Force X             Force Y             Force Z");
    new_line(io);
    putchar(io,'-',65);
    new_line(io);

    new_line(io);

    for glob in 1 .. nn loop
      last := 0;
      for axis in 1 .. 3 loop
        if index(glob,axis) > n and abs force(glob,axis) > maxnorm then
          last := axis;
        end if;
      end loop;

      if last /= 0 then
        put(io,glob,5);
        for axis in 1 .. last loop
          if index(glob,axis) > n and abs force(glob,axis) > maxnorm then
            putchar(io,' ',5);
            put(io,force(glob,axis),2,8,3);
          else
            putchar(io,' ',20);
          end if;
        end loop;
        new_line(io);
      end if;
    end loop;

    close(io);
  end output;

begin
  input;

  start;

  setforce;
  objective(n,f,x,g);

  -- maxnorm
  if mnref <= 0.0 then
    maxnorm := mntol * nlp.norm(n,g);
  else
    maxnorm := mntol * mnref;
  end if;

  major := 0;
  while major < maxmajor and nlp.norm(n,g) > maxnorm loop
    major := major + 1;

    qn.lbfgs(maxlist,n,maxminor,maxls,
             minor,
             maxnorm,lstol,
             f,
             x,g,
             optimal);

    setforce;
    objective(n,f,x,g);
  end loop;

  autocad;

  output;
end orthomem1;
